
/*******************************************************************************
*
* File cdgms.c
*
* Copyright (C) 2014 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* List of all connected diagrams.
*
* The externally accessible functions are
*
*   int cdgms(theory_t th,int order)
*     Generates all connected diagrams in the theory with label "th" up to
*     and including the specified order in the coupling (see the notes).
*     The program returns the total number of diagrams and stores them in
*     static arrays that can be accessed using the other functions in this
*     module.
*
*   int nel_max(theory_t th,int order)
*     Returns the maximal number of external lines of the diagrams listed
*     by cdgms() in the theory with label "th" at the specified order in
*     the coupling.
*
*   int no_of_cdgms(theory_t th,int order,int nel)
*     Returns the number of diagrams with nel external lines generated by
*     cdgms() in the theory with label "th" at the specified order in the
*     coupling.
*
*   dlist_t *select_cdgms(theory_t th,int order,int nel)
*     Returns the pointer to the internally stored structure containing
*     all diagrams with nel external lines generated by cdgms() in the
*     theory with label "th" at the specified order in the coupling. The
*     program returns NULL if there are no such diagrams.
*
*  void iR2D(theory_t th,int *ir,int *id)
*     Computes the indices id of the contracted diagram D in the list
*     generated by cdgms() given the indices ir of the associated labeled
*     tree diagram R in the list generated by ltrees() [trees/ltree_gen.c]
*     in the theory with label th (see the notes).
*
*  void iD2R(theory_t th,int *id,int *ir)
*     Computes the indices ir of the labeled tree diagram R in the list
*     generated by ltrees() [trees/ltree_gen.c] given the indices id of
*     the associated contracted diagram D in the list generated by cdgms()
*     in the theory with label th (see the notes).
*
*   void set_sfactor(theory_t th)
*     Sets the coefficients of the diagrams listed by cdgms() in the theory
*     with label "th" to (-1)^(nq+ng)*Nf^nq/(F1*F2*F3*F4), where nq and ng
*     are the numbers of quark and ghost loops in the diagram and F1*..*F4
*     its total symmetry factor (see diagrams/sfactor.c).
*
* Notes:
*
* The diagrams listed by cdgms() are connected, have no external fermion
* lines and two or more external boson lines. Furthermore, they may not
* have tadpole subdiagrams, i.e. cutting an internal line may not divide
* the diagram into two disconnected parts, where one part has no external
* lines other than the cut line.
*
* Contraction of the labeled lines of the labeled tree diagrams generated
* by ltrees() [trees/ltree_gen.c] yields all diagrams listed by cdgms().
* This mapping of labeled tree diagrams to ordinary diagrams is one-to-one
* and is made explicit by the programs iR2D() and iD2R(). The indices to
* be passed to these programs are:
*
*  id[0]        Order in the coupling of the contracted diagram D.
*
*  id[1]        Number of external boson lines of D.
*
*  id[2]        Index of the diagram in the diagram list returned by
*               select_cdgms(th,id[0],id[1]).
*
*  ir[0]        Order in the coupling of the labeled tree diagram R.
*
*  ir[1]        Number of loop lines of R.
*
*  ir[2]        Index of the diagram in the diagram list returned by
*               select_ltrees(th,ir[0],ir[1]).

* The program cdgms() assumes that trees() [trees/tree_gen.c] and ltrees()
* [trees/ltree_gen.c] have been called before, for the specified theory and
* a sufficiently large order in the coupling. An error occurs if this is
* not the case.
*
* The coefficients of the diagrams are of type cnum_t, where the unknown
* variable is the number Nf of quark flavours if the theory is QCD or LQCD.
* In the case of the other theories, the coefficients are just rational
* numbers (i.e. cnum_t types of degree zero).
*
* The type dlist_t is described at the top of the file dmlist/dmlist.c. The
* lists returned by select_cdgms() include rank and coefficient arrays. All
* stored diagrams have their vertices ranked and ordered by reorder_dgm()
* [diagrams/order.c]. Diagrams that differ by the labeling of the vertices
* and lines are not distinguished.
*
*******************************************************************************/

#define CDGMS_C

#include <stdlib.h>
#include <stdio.h>
#include "utils.h"
#include "contract.h"

typedef struct
{
   int order,*nel;
   int (***id2r)[3],(***ir2d)[3];
   dlist_t ***dl;
} dsets_t;

static int nrat=0;
static long (*rat)[2];
static vlist_t vls;
static dsets_t cds[(int)(THEORIES)]={{0,NULL,NULL,NULL,NULL},
                                     {0,NULL,NULL,NULL,NULL},
                                     {0,NULL,NULL,NULL,NULL},
                                     {0,NULL,NULL,NULL,NULL},
                                     {0,NULL,NULL,NULL,NULL}};


static void alloc_dl(theory_t th,int order)
{
   int i,j,n,*nel;
   dlist_t ***dl,**dll;
   dsets_t *cd;

   nel=malloc((order+1)*sizeof(*nel));
   error(nel==NULL,1,"alloc_dl [cdgms.c]",
         "Unable to allocate diagram set structure");
   n=0;
   nel[0]=0;

   for (i=1;i<=order;i++)
   {
      if (th==PHI4)
         nel[i]=2*i+2;
      else
         nel[i]=i+2;

      n+=(nel[i]+1);
   }

   dl=malloc((order+1)*sizeof(*dl));
   dll=malloc(n*sizeof(*dll));
   error((dl==NULL)||(dll==NULL),2,"alloc_dl [cdgms.c]",
         "Unable to allocate diagram set structure");
   dl[0]=NULL;

   for (i=1;i<=order;i++)
   {
      dl[i]=dll;

      for (j=0;j<=nel[i];j++)
      {
         if (j>=2)
            dll[j]=alloc_dlist(32,3);
         else
            dll[j]=NULL;
      }

      dll+=(nel[i]+1);
   }

   cd=cds+(int)(th);
   (*cd).order=order;
   (*cd).nel=nel;
   (*cd).dl=dl;
}


static int get_nel(theory_t th,ltree_t *ltd)
{
   int nv,nl,nel;
   int i,(*v)[2];
   ltype_t lt;
   line_t *l;
   vertex_t *vx;
   diagram_t *dm;

   if (th==PHI4)
      lt=SCALAR;
   else
      lt=GLUON;

   dm=(*ltd).dm;
   nv=(*dm).nv;
   v=(*dm).v;
   nel=0;

   for (i=0;i<nv;i++)
   {
      vx=vls.v[v[0][0]]+v[0][1];
      nel+=(*vx).nl[(int)(lt)];
      v+=1;
   }

   nl=(*dm).nl;
   l=(*dm).l;

   for (i=0;i<nl;i++)
   {
      if ((*l).type==lt)
         nel-=2;
      l+=1;
   }

   nl=(*ltd).nl;
   l=(*ltd).l;

   for (i=0;i<nl;i++)
   {
      if ((*l).type==lt)
         nel-=2;
      l+=1;
   }

   return nel;
}


static void set_dl(theory_t th,int order)
{
   int nlp,nel,ie;
   int n,i,j,k,*rkn;
   diagram_t *dm,*dmn;
   ltree_t **lta,*ltd;
   tlist_t *td;
   dsets_t *cd;

   ie=1;
   cd=cds+(int)(th);

   for (i=1;i<=order;i++)
   {
      nlp=nlp_max(th,i);

      for (j=0;j<=nlp;j++)
      {
         td=select_ltrees(th,i,j);

         if (td!=NULL)
         {
            n=(*td).n;
            lta=(*td).lta;

            for (k=0;k<n;k++)
            {
               ltd=lta[k];
               dm=(*ltd).dm;
               dmn=alloc_dgm((*dm).nv,(*dm).nl+(*ltd).nl);
               rkn=alloc_rank((*dm).nv);
               ltd2dgm(ltd,dmn);
               reorder_dgm(dmn,rkn);
               nel=get_nel(th,ltd);

               if ((nel>=2)&&(nel<=(*cd).nel[i]))
                  ie&=insert_dgm(dmn,rkn,(*cd).dl[i][nel]);
               else
                  error(1,1,"set_dl [cdgms.c]",
                        "Unexpected number of external lines");
            }
         }
      }
   }

   error(ie!=1,1,"set_dl [cdgms.c]","Duplicate labeled tree diagrams");
}


static void alloc_id2r(theory_t th,int order)
{
   int i,j,*nel;
   int n1,n2,n3;
   int (***id1)[3],(**id2)[3],(*id3)[3];
   dlist_t *dl;
   dsets_t *cd;

   cd=cds+(int)(th);
   nel=(*cd).nel;
   n1=order+1;
   n2=0;
   n3=0;

   for (i=1;i<=order;i++)
   {
      n2+=(nel[i]+1);

      for (j=2;j<=nel[i];j++)
      {
         dl=(*cd).dl[i][j];
         n3+=(*dl).n;
      }
   }

   id1=malloc(n1*sizeof(*id1));
   id2=malloc(n2*sizeof(*id2));
   id3=malloc(n3*sizeof(*id3));

   error((id1==NULL)||(id2==NULL)||(id3==NULL),1,"alloc_id2r [cdgms.c]",
         "Unable to allocate index array");

   id1[0]=NULL;

   for (i=1;i<=order;i++)
   {
      id1[i]=id2;

      for (j=0;j<=nel[i];j++)
      {
         dl=(*cd).dl[i][j];

         if ((j>=2)&&((*dl).n!=0))
         {
            id2[j]=id3;
            id3+=(*dl).n;
         }
         else
            id2[j]=NULL;
      }

      id2+=(nel[i]+1);
   }

   (*cd).id2r=id1;
}


static void alloc_ir2d(theory_t th,int order)
{
   int i,j,n,nlp;
   int n1,n2,n3;
   int (***ir1)[3],(**ir2)[3],(*ir3)[3];
   dsets_t *cd;

   n1=order+1;
   n2=0;
   n3=0;

   for (i=1;i<=order;i++)
   {
      nlp=nlp_max(th,i);
      n2+=(nlp+1);

      for (j=0;j<=nlp;j++)
         n3+=no_of_ltrees(th,i,j);
   }

   ir1=malloc(n1*sizeof(*ir1));
   ir2=malloc(n2*sizeof(*ir2));
   ir3=malloc(n3*sizeof(*ir3));

   error((ir1==NULL)||(ir2==NULL)||(ir3==NULL),1,
         "alloc_ir2d [cdgms.c]","Unable to allocate index array");

   ir1[0]=NULL;

   for (i=1;i<=order;i++)
   {
      nlp=nlp_max(th,i);
      ir1[i]=ir2;

      for (j=0;j<=nlp;j++)
      {
         n=no_of_ltrees(th,i,j);

         if (n>0)
         {
            ir2[j]=ir3;
            ir3+=n;
         }
         else
            ir2[j]=NULL;
      }

      ir2+=(nlp+1);
   }

   cd=cds+(int)(th);
   (*cd).ir2d=ir1;
}


static void set_indices(theory_t th,int order)
{
   int nlp,nel;
   int nt,nd,i,j,k;
   int (***ir2d)[3],(***id2r)[3];
   ltree_t **lta,*ltd;
   tlist_t *td;
   dlist_t *dl;
   dsets_t *cd;

   cd=cds+(int)(th);
   ir2d=(*cd).ir2d;
   id2r=(*cd).id2r;

   for (i=1;i<=order;i++)
   {
      nel=(*cd).nel[i];

      for (j=2;j<=nel;j++)
      {
         dl=(*cd).dl[i][j];
         (*dl).n=0;
      }
   }

   for (i=1;i<=order;i++)
   {
      nlp=nlp_max(th,i);

      for (j=0;j<=nlp;j++)
      {
         td=select_ltrees(th,i,j);

         if (td!=NULL)
         {
            nt=(*td).n;
            lta=(*td).lta;

            for (k=0;k<nt;k++)
            {
               ltd=lta[k];
               nel=get_nel(th,ltd);
               dl=(*cd).dl[i][nel];
               nd=(*dl).n;

               ir2d[i][j][k][0]=i;
               ir2d[i][j][k][1]=nel;
               ir2d[i][j][k][2]=nd;

               id2r[i][nel][nd][0]=i;
               id2r[i][nel][nd][1]=j;
               id2r[i][nel][nd][2]=k;

               (*dl).n+=1;
            }
         }
      }
   }
}


static int nall_dgm(theory_t th,int order)
{
   int nall,i,j;
   int *nel;
   dlist_t *dl;
   dsets_t *cd;

   cd=cds+(int)(th);
   nel=(*cd).nel;
   nall=0;

   for (i=1;i<=order;i++)
   {
      for (j=2;j<=nel[i];j++)
      {
         dl=(*cd).dl[i][j];
         nall+=(*dl).n;
      }
   }

   return nall;
}


int cdgms(theory_t th,int order)
{
   dsets_t *cd;

   error((th==THEORIES)||(order<0),1,"cdgms [cdgms.c]",
         "Arguments are out of range");
   cd=cds+(int)(th);

   if ((order>0)&&((*cd).order==0))
   {
      vls=vertices(th,order);
      alloc_dl(th,order);
      set_dl(th,order);
      alloc_id2r(th,order);
      alloc_ir2d(th,order);
      set_indices(th,order);
   }
   else if (order>(*cd).order)
      error(1,1,"cdgms [cdgms.c]","Attempt to extend the diagram list");

   if (order>0)
      return nall_dgm(th,order);
   else
      return 0;
}


int nel_max(theory_t th,int order)
{
   dsets_t *cd;

   error((th==THEORIES)||(order<0),1,"nel_max [cdgms.c]",
         "Arguments are out of range");

   cd=cds+(int)(th);
   error(order>(*cd).order,1,"nel_max [cdgms.c]",
         "Requested diagrams are not included in current list");

   return (*cd).nel[order];
}


int no_of_cdgms(theory_t th,int order,int nel)
{
   dlist_t *dl;
   dsets_t *cd;

   error((th==THEORIES)||(order<0),1,"no_of_cdgms [cdgms.c]",
         "Arguments are out of range");

   if (order==0)
      return 0;

   cd=cds+(int)(th);
   error(order>(*cd).order,1,"no_of_cdgms [cdgms.c]",
         "Requested diagrams are not included in current list");

   error((nel<0)||(nel>(*cd).nel[order]),1,"no_of_cdgms [cdgms.c]",
         "Parameter nel is out of range");

   if (nel<2)
      return 0;

   dl=(*cd).dl[order][nel];

   return (*dl).n;
}


dlist_t *select_cdgms(theory_t th,int order,int nel)
{
   dlist_t *dl;
   dsets_t *cd;

   error((th==THEORIES)||(order<0),1,"select_cdgms [cdgms.c]",
         "Arguments are out of range");

   if (order==0)
      return NULL;

   cd=cds+(int)(th);
   error(order>(*cd).order,1,"select_cdgms [cdgms.c]",
         "Requested diagrams are not included in current list");

   error((nel<0)||(nel>(*cd).nel[order]),1,"select_cdgms [cdgms.c]",
         "Parameter nel is out of range");

   if (nel<2)
      return NULL;

   dl=(*cd).dl[order][nel];

   if ((*dl).n==0)
      return NULL;
   else
      return dl;
}


void iR2D(theory_t th,int *ir,int *id)
{
   int *ir2d;
   dsets_t *cd;

   error((ir[2]<0)||(ir[2]>=no_of_ltrees(th,ir[0],ir[1])),1,
         "iR2D [cdgms.c]","Input index ir[2] is out of range");

   cd=cds+(int)(th);
   ir2d=(*cd).ir2d[ir[0]][ir[1]][ir[2]];
   id[0]=ir2d[0];
   id[1]=ir2d[1];
   id[2]=ir2d[2];
}


void iD2R(theory_t th,int *id,int *ir)
{
   int *id2r;
   dsets_t *cd;

   error((id[2]<0)||(id[2]>=no_of_cdgms(th,id[0],id[1])),1,
         "iD2R [cdgms.c]","Input index id[2] is out of range");

   cd=cds+(int)(th);
   id2r=(*cd).id2r[id[0]][id[1]][id[2]];
   ir[0]=id2r[0];
   ir[1]=id2r[1];
   ir[2]=id2r[2];
}


static void set_rat(int p,long f)
{
   int i;

   if (p>=nrat)
   {
      if (nrat>0)
         free(rat);

      nrat=p+1;
      rat=malloc(nrat*sizeof(*rat));
      error(rat==NULL,1,"set_rat [cdgms.c]",
            "Unable to allocate rational-number array");
   }

   for (i=0;i<p;i++)
   {
      rat[i][0]=0L;
      rat[i][1]=1L;
   }

   rat[p][0]=1L;
   rat[p][1]=f;
}


void set_sfactor(theory_t th)
{
   int order,*nel;
   int nq,ng,n,i,j,k;
   long f,sf[4];
   cnum_t **cna;
   diagram_t **dma;
   dlist_t *dl;
   dsets_t *cd;

   error(th==THEORIES,1,"set_sfactor [cdgms.c]",
         "Argument is out of range");

   cd=cds+(int)(th);
   order=(*cd).order;
   nel=(*cd).nel;

   for (i=1;i<=order;i++)
   {
      for (j=2;j<=nel[i];j++)
      {
         dl=(*cd).dl[i][j];
         n=(*dl).n;
         dma=(*dl).dma;
         cna=(*dl).cna;

         for (k=0;k<n;k++)
         {
            if ((th==QCD)||(th==LQCD))
               nq=nquark_loops(dma[k]);
            else
               nq=0;

            if (th!=PHI4)
               ng=nghost_loops(dma[k]);
            else
               ng=0;

            f=sfactor(th,dma[k],sf);

            if ((nq+ng)&0x1)
               f=-f;

            set_rat(nq,f);
            set_cnum(nq,rat,cna[k]);
         }
      }
   }
}
