
/*******************************************************************************
*
* File tmap.c
*
* Copyright (C) 2014 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Contruction of the trivializing map.
*
* The externally accessible function is
*
*   int tmap(theory_t th,int order,int is)
*     Generates the rooted tree diagrams contributing to the trivializing
*     map in the theory with label "th" up to the specified order in the
*     coupling and returns the number of these diagrams. The coefficients
*     of the diagrams are computed as well and stored together with the
*     diagrams (see the notes).
*
* Notes:
*
* The program tmap() calls trees() and ltrees() to generate the labeled tree
* diagrams for the chosen theory. The root lines are selected by calling
* set_root(theory_t th,int is) [trees/ltree_gen.c].
*
* Once the diagrams are generated, their coefficients are calculated following
* the lines of section 7 of the paper
*
*  M. Luescher: "Instantaneous stochastic perturbation theory"
*
* The diagrams and their coefficients are stored in statically allocated
* tlist_t [trees.h] structures, whose contents can be accessed using the
* program select_ltrees() [trees/ltree_gen.c].
*
* The program tmap() generates all rooted labeled tree diagrams that are
* expected to contribute to the trivializing map. However, the coefficients
* some diagrams may vanish accidentally.
*
*******************************************************************************/

#define TMAP_C

#include <stdlib.h>
#include <stdio.h>
#include "utils.h"
#include "permutations.h"
#include "contract.h"

static int init=0;
static int ml,mj,cidl,cidj;
static int *ki,*li,*ji;
static int *nl,*nj,(*rl)[2],(*rj)[2];
static int nvw,nlw,*rkw;
static long nck,ncl,ncj;
static long nmk,nml,nmj;
static cnum_t *cns;
static line_t *lnl;
static vlist_t vls;
static diagram_t *dmw;
static vsets_t vs;
static src_legs_t sl;
static ltree_t **ltd;
static dlist_t ***dls;
static tlist_t ***tls;


static void set_dls(theory_t th,int order)
{
   int i,j;
   int n1,n2,nel;
   dlist_t ***dl1,**dl2;

   n1=order+1;
   n2=0;

   for (i=1;i<=order;i++)
      n2+=(nel_max(th,i)+1);

   dl1=malloc(n1*sizeof(*dl1));
   dl2=malloc(n2*sizeof(*dl2));
   error((dl1==NULL)||(dl2==NULL),1,"set_dls [tmap.c]",
         "Unable to allocate dlist array");
   dl1[0]=NULL;

   for (i=1;i<=order;i++)
   {
      dl1[i]=dl2;
      nel=nel_max(th,i);

      for (j=0;j<=nel;j++)
      {
         if (j>=2)
            dl2[j]=select_cdgms(th,i,j);
         else
            dl2[j]=NULL;
      }

      dl2+=(nel+1);
   }

   dls=dl1;
}


static void free_dls(void)
{
   free(dls[1]);
   free(dls);
}


static void set_tls(theory_t th,int order)
{
   int i,j;
   int n1,n2,nlp;
   tlist_t ***tl1,**tl2;

   n1=order+1;
   n2=0;

   for (i=1;i<=order;i++)
      n2+=(nlp_max(th,i)+1);

   tl1=malloc(n1*sizeof(*tl1));
   tl2=malloc(n2*sizeof(*tl2));
   error((tl1==NULL)||(tl2==NULL),1,"set_tls [tmap.c]",
         "Unable to allocate tlist array");
   tl1[0]=NULL;

   for (i=1;i<=order;i++)
   {
      tl1[i]=tl2;
      nlp=nlp_max(th,i);

      for (j=0;j<=nlp;j++)
         tl2[j]=select_ltrees(th,i,j);

      tl2+=(nlp+1);
   }

   tls=tl1;
}


static void free_tls(void)
{
   free(tls[1]);
   free(tls);
}


static void alloc_dmw(theory_t th,int order)
{
   int i,j,k;
   int n,nel;
   diagram_t *dm,**dma;
   dlist_t *dl;

   nvw=0;
   nlw=0;

   for (i=1;i<=order;i++)
   {
      nel=nel_max(th,i);

      for (j=2;j<=nel;j++)
      {
         dl=dls[i][j];

         if (dl!=NULL)
         {
            n=(*dl).n;
            dma=(*dl).dma;

            for (k=0;k<n;k++)
            {
               dm=dma[k];

               if (nvw<(*dm).nv)
                  nvw=(*dm).nv;

               if (nlw<(*dm).nl)
                  nlw=(*dm).nl;
            }
         }
      }
   }

   dmw=alloc_dgm(nvw,nlw);
   rkw=alloc_rank(nvw);

   if (nlw>0)
   {
      lnl=malloc(nlw*sizeof(*lnl));
      error(lnl==NULL,1,"alloc_dmw [tmap.c]","Unable to allocate line list");
   }
   else
      lnl=NULL;
}


static void free_dmw(void)
{
   free_dgm(dmw);
   free(rkw);

   if (nlw>0)
      free(lnl);
}


static void alloc_aux(int n)
{
   ki=malloc(5*n*sizeof(*ki));
   rl=malloc(2*n*sizeof(*rl));
   ltd=malloc(n*sizeof(*ltd));

   error((ki==NULL)||(rl==NULL)||(ltd==NULL),1,
         "alloc_aux [tmap.c]","Unable to allocate auxiliary arrays");

   li=ki+n;
   ji=li+n;
   nl=ji+n;
   nj=nl+n;
   rj=rl+n;
}


static void free_aux(void)
{
   free(ltd);
   free(rl);
   free(ki);
}


static void dgm_type(theory_t th,int *id)
{
   int n,i,(*v)[2];
   ltype_t lt;
   line_t *l;
   vertex_t *vx;

   if (th==PHI4)
      lt=SCALAR;
   else
      lt=GLUON;

   n=(*dmw).nv;
   v=(*dmw).v;
   id[0]=0;
   id[1]=0;

   for (i=0;i<n;i++)
   {
      vx=vls.v[v[0][0]]+v[0][1];
      id[0]+=v[0][0];
      id[1]+=(*vx).nl[(int)(lt)];
      v+=1;
   }

   n=(*dmw).nl;
   l=(*dmw).l;

   for (i=0;i<n;i++)
   {
      if ((*l).type==lt)
         id[1]-=2;
      l+=1;
   }
}


static void find_dgm(theory_t th,int *id)
{
   dlist_t *dl;

   reorder_dgm(dmw,rkw);
   dgm_type(th,id);
   dl=dls[id[0]][id[1]];

   if (dl!=NULL)
   {
      id[2]=search_dgm(dmw,rkw,dl);

      if (id[2]<(*dl).n)
         return;
   }

   error(1,1,"find_dgm [tmap.c]","Diagram not found");
}


static void init_coeffs(theory_t th,int order)
{
   int n,i,j,k;
   int nlp,ir[3],id[3];
   cnum_t **cna;
   dlist_t *dl;
   tlist_t *tl;

   for (i=1;i<=order;i++)
   {
      ir[0]=i;
      nlp=nlp_max(th,i);

      for (j=0;j<=nlp;j++)
      {
         ir[1]=j;
         tl=tls[i][j];

         if (tl!=NULL)
         {
            n=(*tl).n;
            cna=(*tl).cna;

            for (k=0;k<n;k++)
            {
               ir[2]=k;
               iR2D(th,ir,id);
               dl=dls[id[0]][id[1]];
               copy_cnum((*dl).cna[id[2]],cna[k]);
            }
         }
      }
   }
}


static void set_nl(int n)
{
   int i;

   ml=0;
   nl[0]=1;

   for (i=1;i<n;i++)
   {
      if (ki[i]==ki[i-1])
         nl[ml]+=1;
      else
      {
         ml+=1;
         nl[ml]=1;
      }
   }

   ml+=1;
}


static void set_rl(theory_t th)
{
   int i,j;

   j=0;

   for (i=0;i<ml;i++)
   {
      rl[i][0]=0;
      rl[i][1]=nlp_max(th,ki[j]);
      j+=nl[i];
   }
}


static void set_nj(int n)
{
   int i;

   mj=0;
   nj[0]=1;

   for (i=1;i<n;i++)
   {
      if ((ki[i]==ki[i-1])&&(li[i]==li[i-1]))
         nj[mj]+=1;
      else
      {
         mj+=1;
         nj[mj]=1;
      }
   }

   mj+=1;
}


static int set_rj(theory_t th)
{
   int i,j;
   tlist_t *tl;

   j=0;

   for (i=0;i<mj;i++)
   {
      rj[i][0]=0;
      tl=tls[ki[j]][li[j]];

      if (tl!=NULL)
         rj[i][1]=(*tl).n-1;
      else
         return 0;

      j+=nj[i];
   }

   return 1;
}


static void set_ltd(int n)
{
   int i;
   long r[1][2];
   cnum_t **cna;
   ltree_t **lta;
   tlist_t *tl;

   r[0][0]=1L;
   r[0][1]=1L;
   set_cnum(0,r,cns+1);

   for (i=0;i<n;i++)
   {
      tl=tls[ki[i]][li[i]];
      lta=(*tl).lta;
      cna=(*tl).cna;

      ltd[i]=lta[ji[i]];
      mul_cnum(cna[ji[i]],cns+1);
   }
}


static void contract0(theory_t th,int n)
{
   int i,nmx,nln;
   int ir[3],id[3];
   long np,npm,r[1][2];
   ltype_t lt;
   tlist_t *tl;

   if (th==PHI4)
      lt=SCALAR;
   else
      lt=GLUON;

   error((sl.nt<1)||(sl.lt[0]!=lt)||(sl.lbl[0]!=0)||(sl.ns[0]<2),1,
         "contract0 [tmap.c]","Unexpected source inventory");

   nmx=sl.ns[0]/2;
   nln=0;

   for (i=1;i<sl.nt;i++)
      nln+=sl.ns[i];

   nln/=2;
   r[0][1]=mfact_ltrees(mj,nj,ji);

   for (i=0;i<=nmx;i++)
   {
      npm=init_contract(i,&sl);
      np=0L;

      while (np<npm)
      {
         np=next_contract(lnl,r[0]);

         if (lnlcon(&vs,nln+i,lnl)==1)
         {
            (*dmw).nv=nvw;
            (*dmw).nl=nlw;
            merge_ltrees(n,ltd,nln+i,lnl,dmw);

            if (has_tadpole(dmw,lt,&vls)==0)
            {
               find_dgm(th,id);
               iD2R(th,id,ir);
               tl=tls[ir[0]][ir[1]];

               set_cnum(0,r,cns);
               mul_cnum(cns+1,cns);
               sub_cnum(cns,(*tl).cna[ir[2]]);
            }
         }
      }

      stop_contract();
   }
}


static void contract1(theory_t th)
{
   int i,nmx,nln;
   int ir[3],id[3];
   long np,npm,r[1][2];
   ltype_t lt;
   tlist_t *tl;

   if (th==PHI4)
      lt=SCALAR;
   else
      lt=GLUON;

   error((sl.nt<1)||(sl.lt[0]!=lt)||(sl.lbl[0]!=0)||(sl.ns[0]<1),1,
         "contract1 [tmap.c]","Unexpected source inventory");

   nmx=(sl.ns[0]-1)/2;
   nln=0;

   for (i=1;i<sl.nt;i++)
      nln+=sl.ns[i];

   nln/=2;
   r[0][1]=1L;

   for (i=1;i<=nmx;i++)
   {
      npm=init_contract(i,&sl);
      np=0L;

      while (np<npm)
      {
         np=next_contract(lnl,r[0]);

         (*dmw).nv=nvw;
         (*dmw).nl=nlw;
         merge_ltrees(1,ltd,nln+i,lnl,dmw);

         if (has_tadpole(dmw,lt,&vls)==0)
         {
            find_dgm(th,id);
            iD2R(th,id,ir);
            tl=tls[ir[0]][ir[1]];

            set_cnum(0,r,cns);
            mul_cnum(cns+1,cns);
            sub_cnum(cns,(*tl).cna[ir[2]]);
         }
      }

      stop_contract();
   }
}


static void subtract0(theory_t th,int k)
{
   int n;

   for (n=2;n<=k;n++)
   {
      nmk=init_partition(n,k);
      nck=0L;

      while (nck<nmk)
      {
         nck=next_partition(ki);
         set_nl(n);
         set_rl(th);
         nml=init_mindex(ml,nl,rl,&cidl);
         ncl=0L;

         while (ncl<nml)
         {
            ncl=next_mindex(cidl,li);
            set_nj(n);

            if (set_rj(th))
            {
               nmj=init_mindex(mj,nj,rj,&cidj);
               ncj=0L;

               while (ncj<nmj)
               {
                  ncj=next_mindex(cidj,ji);
                  set_ltd(n);
                  get_vsets(n,ltd,&vs);
                  get_src_legs(th,n,ltd,&sl);
                  contract0(th,n);
               }

               stop_mindex(cidj);
            }
         }

         stop_mindex(cidl);
      }

      stop_partition();
   }
}


static void subtract1(theory_t th,int k,int l)
{
   int n,i;
   cnum_t **cna;
   ltree_t **lta;
   tlist_t *tl;

   tl=tls[k][l];

   if (tl!=NULL)
   {
      n=(*tl).n;
      lta=(*tl).lta;
      cna=(*tl).cna;

      for (i=0;i<n;i++)
      {
         ltd[0]=lta[i];
         get_src_legs(th,1,ltd,&sl);
         copy_cnum(cna[i],cns+1);
         contract1(th);
      }
   }
}


int tmap(theory_t th,int order,int is)
{
   int n,k,l,nlp;

   if (order==0)
      return 0;

   if (init==0)
   {
      cns=alloc_cnum(2,0);
      init_vsets(&vs);
      init_src_legs(&sl);
      init=1;
   }

   vls=vertices(th,order);
   (void)(trees(th,order));
   n=ltrees(th,order);
   set_root(th,is);
   (void)(cdgms(th,order));
   set_sfactor(th);

   set_dls(th,order);
   set_tls(th,order);
   alloc_dmw(th,order);
   alloc_aux(order);
   init_coeffs(th,order);

   for (k=1;k<=order;k++)
   {
      subtract0(th,k);
      nlp=nlp_max(th,k);

      for (l=0;l<=nlp;l++)
         subtract1(th,k,l);
   }

   free_aux();
   free_dmw();
   free_tls();
   free_dls();

   return n;
}
