
/*******************************************************************************
*
* File tdlist.c
*
* Copyright (C) 2014 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Utility programs for lists of labeled tree diagrams.
*
* The externally accessible functions are
*
*   tlist_t *alloc_tlist(int nmax)
*     Creates a new diagram list, with initially space for nmax diagrams.
*     The program returns the pointer to the new list.
*
*   void free_tlist(tlist_t *tl)
*     Frees the arrays in the diagram list tl and then the structure
*     itself. The diagrams pointed to by the diagram array and the
*     coefficients pointed to by the coefficient array are not freed.
*     The program does nothing if tl=NULL.
*
*   void deep_free_tlist(tlist_t *tl)
*     Frees the diagrams and coefficients pointed to by the diagram and
*     coefficient arrays in the diagram list tl and then the list itself.
*     The program does nothing if tl=NULL.
*
*   void add_ltd(ltree_t *td,tlist_t *tl)
*     Adds the labeled tree diagram td to the diagram list tl and sets its
*     coefficient to 0. The diagram structure td must have been allocated
*     using alloc_ltd() [ltree.c].
*
* Notes:
*
* The structures of type tlist_t have the following elements:
*
*  n                     Current number of diagrams included in the
*                        list.
*
*  nmax                  Current maximal number of diagrams that can
*                        be included in the list. nmax and the arrays in
*                        the list are adjusted dynamically by add_ltd().
*
*  lta[i]                Pointer to the labeled tree diagram number i
*                        (i=0,..,n-1).
*
*  cna[i]                Pointer to the coefficient (of type cnum_t) of
*                        the i'th diagram.
*
* Labeled tree diagrams are described by structures of type ltree_t (see
* ltree.c). Unused elements of the diagram array and the coefficient array
* are set to NULL.
*
* The type cnum_t is defined in utils.h and programs manipulating numbers
* of this type are collected in utils/cnum.c.
*
* Note that the array addresses (*tl).lta and (*tl).cna may change when a
* diagram is added to the diagram list tl.
*
*******************************************************************************/

#define TDLIST_C

#include <stdlib.h>
#include <stdio.h>
#include <limits.h>
#include "utils.h"
#include "dmlist.h"

static long r0[1][2]={{0L,1L}};


static ltree_t **new_lta(int nmax,int nold,ltree_t **ltold)
{
   int i;
   ltree_t **lta;

   lta=malloc(nmax*sizeof(*lta));
   error(lta==NULL,1,"new_lta [tdlist.c]",
         "Unable to allocate diagram array");

   for (i=0;i<nold;i++)
      lta[i]=ltold[i];

   for (i=nold;i<nmax;i++)
      lta[i]=NULL;

   if (nold>0)
      free(ltold);

   return lta;
}


static cnum_t **new_cna(int nmax,int nold,cnum_t **cnold)
{
   int i;
   cnum_t **cna;

   cna=malloc(nmax*sizeof(*cna));
   error(cna==NULL,1,"new_cna [tdlist.c]",
         "Unable to allocate coefficient array");

   for (i=0;i<nold;i++)
      cna[i]=cnold[i];

   for (i=nold;i<nmax;i++)
      cna[i]=NULL;

   if (nold>0)
      free(cnold);

   return cna;
}


tlist_t *alloc_tlist(int nmax)
{
   tlist_t *tl;

   error(nmax<1,1,"alloc_tlist [tdlist.c]",
         "Argument nmax is out of range");

   tl=malloc(1*sizeof(*tl));
   error(tl==NULL,1,"alloc_tlist [tdlist.c]",
         "Unable to allocate diagram list");

   (*tl).n=0;
   (*tl).nmax=nmax;
   (*tl).lta=new_lta(nmax,0,NULL);
   (*tl).cna=new_cna(nmax,0,NULL);

   return tl;
}


void free_tlist(tlist_t *tl)
{
   if (tl!=NULL)
   {
      free((*tl).lta);
      free((*tl).cna);
      free(tl);
   }
}


void deep_free_tlist(tlist_t *tl)
{
   int n,i;
   cnum_t **cna;
   ltree_t **lta;

   if (tl!=NULL)
   {
      n=(*tl).n;
      lta=(*tl).lta;
      cna=(*tl).cna;

      for (i=0;i<n;i++)
      {
         free_ltd(lta[i]);
         free_cnum(1,cna[i]);
      }

      free_tlist(tl);
   }
}


void add_ltd(ltree_t *td,tlist_t *tl)
{
   int m,n,nmax;

   n=(*tl).n;
   nmax=(*tl).nmax;

   if (n==nmax)
   {
      if (nmax<=(INT_MAX/2))
         m=2*nmax;
      else
      {
         error(nmax==INT_MAX,1,"add_ltd [tdlist.c]",
               "Unable to extend diagram list");
         m=INT_MAX;
      }

      (*tl).nmax=m;
      (*tl).lta=new_lta(m,nmax,(*tl).lta);
      (*tl).cna=new_cna(m,nmax,(*tl).cna);
   }

   (*tl).n=n+1;
   (*tl).lta[n]=td;

   if ((*tl).cna[n]==NULL)
      (*tl).cna[n]=alloc_cnum(1,0);
   else
      set_cnum(0,r0,(*tl).cna[n]);
}
