
/*******************************************************************************
*
* File branch.c
*
* Copyright (C) 2014 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Generation and manipulation of branch diagrams.
*
* The externally accessible functions are
*
*   branch_t *alloc_branch(int nv)
*     Returns the pointer to a newly allocated branch structure for nv
*     vertices with properly allocated and initialized arrays (integers
*     to 0, line types to LTYPES).
*
*   void free_branch(branch_t *br)
*     Frees the arrays in the branch structure br and then the structure
*     itself.
*
*   int branches(theory_t th,int order)
*     Generates a tree of linked vertices that contains all branches in the
*     theory with label "th" up to and including the specified order in the
*     coupling. The program returns the total number of branches.
*
*   int get_nv(theory_t th,int ibr)
*     Returns the number of vertices of the branch with index ibr in the
*     tree of linked vertices built by branches() for the theory with label
*     th.
*
*   int get_order(theory_t th,int ibr)
*     Returns the order in the coupling of the branch with index ibr in
*     the tree of linked vertices built by branches() for the theory with
*     label th.
*
*   void get_branch(theory_t th,int ibr,branch_t *br)
*     Retrieves the branch number ibr from the tree of linked vertices built
*     by branches() for the theory with label th and assigns the branch
*     elements to the structure br. The arrays in the latter are assumed to
*     be allocated with sufficiently large size (an error occurs if not).
*
*   int find_branch(theory_t th,branch_t *br)
*     Returns the index ibr of the branch br in the tree of linked vertices
*     built by branches() for the theory with label th. An error occurs if
*     the branch is not found.
*
*   int inverse_branch(theory_t th,int ibr)
*     Returns the index of the branch obtained by inverting the order of the
*     vertices and lines of the branch with index ibr in the tree of linked
*     vertices built by branches(th,..). The inversion of the order of the
*     lines includes conjugating the line type so that the diagram remains
*     the same.
*
*   void branch2dgm(branch_t *br,diagram_t *dm)
*     Converts the branch br to the diagram dm. The arrays in the latter are
*     assumed to be allocated with sufficiently large size (an error occurs
*     if not). The conversion preserves the ordering of the vertices and lines
*     in the branch.
*
* Notes:
*
* Currently the supported theories and their name tags are:
*
*  PHI4                  Phi^4 theory.
*
*  YM                    Continuum pure gauge theory.
*
*  QCD                   Continuum QCD.
*
*  LYM                   Pure lattice gauge theory.
*
*  LQCD                  Lattice QCD.
*
* A branch is an ordered sequence of vertices and lines. All vertices must be
* of type ACTION,MASS,MEASURE,GFIX or SOURCE, where only the first and the last
* vertex may be of type SOURCE. The lines connect subsequent vertices, the line
* number i going out of the vertex number i.
*
* The type branch_t is a structure with the following elements:
*
*  nv                    Number of vertices.
*
*  v[i][0],v[i][1]       Indices k,j of the vertex number i in the list
*                        of vertices of the theory (i=0,..,nv-1).
*
*  l[i]                  Type of the connecting line number i (i=0,..,nv-2).
*
* Lines are an enum type with values SCALAR,GLUON,GHOST,ANTIGHOST,QUARK
* and ANTIQUARK. At the vertices, ingoing directed lines are counted as
* outgoing conjugate lines (GHOST <-> ANTIGHOST, QUARK <-> ANTIQUARK).
* An outgoing quark line counts as QUARK, for example, and so does an
* ingoing antiquark line.
*
* The program branches() constructs a tree of linked vertices that contains
* all branches up to and including the specified order in the coupling. Each
* branch is assigned a unique index ibr=0,1,..,nbr-1, where nbr is the total
* number of branches. Once constructed, the tree of linked vertices cannot
* be modified by calling branches() with different values of the parameter
* "order". If called with the same value of "order", the program returns the
* total number of branches included in the tree.
*
*******************************************************************************/

#define BRANCH_C

#include <stdlib.h>
#include <stdio.h>
#include "utils.h"
#include "rvt.h"
#include "trees.h"

typedef struct bvertex_t
{
   int order,ibr,nup;
   int v[2];
   ltype_t l;
   struct bvertex_t *up,*down;
} bvertex_t;

typedef struct
{
   int order,nbr,nup;
   bvertex_t *up;
} blist_t;

static int *binv[(int)(THEORIES)]={NULL,NULL,NULL,NULL,NULL};
static bvertex_t **btag[(int)(THEORIES)]={NULL,NULL,NULL,NULL,NULL};
static blist_t bls[(int)(THEORIES)]={{0,0,0,NULL},{0,0,0,NULL},{0,0,0,NULL},
                                     {0,0,0,NULL},{0,0,0,NULL}};


branch_t *alloc_branch(int nv)
{
   int i,(*v)[2];
   ltype_t *l;
   branch_t *br;

   error(nv<1,1,"alloc_branch [branch.c]",
         "Branches must contain at least 1 vertex");

   br=malloc(1*sizeof(*br));
   v=malloc(nv*sizeof(*v));

   if (nv>1)
      l=malloc((nv-1)*sizeof(*l));
   else
      l=NULL;

   error((br==NULL)||(v==NULL)||((nv>1)&&(l==NULL)),1,
         "alloc_branch [branch.c]","Unable to allocate branch");

   for (i=0;i<nv;i++)
   {
      v[i][0]=0;
      v[i][1]=0;

      if (i<(nv-1))
         l[i]=LTYPES;
   }

   (*br).nv=nv;
   (*br).v=v;
   (*br).l=l;

   return br;
}


void free_branch(branch_t *br)
{
   if (br!=NULL)
   {
      free((*br).v);

      if ((*br).nv>1)
         free((*br).l);

      free(br);
   }
}


static int add_vertices(int order,int nbr,vlist_t *vl,bvertex_t *bv)
{
   int nup,imx,i,j,k;
   ltype_t l,ln;
   vertex_t *v,*vn;
   bvertex_t *up;

   v=(*vl).v[(*bv).v[0]]+(*bv).v[1];
   imx=order-(*bv).order;
   if (imx>(*vl).order)
      imx=(*vl).order;
   l=rvt[(*bv).l];
   nup=0;

   for (i=0;i<=imx;i++)
   {
      for (j=0;j<(*vl).nv[i];j++)
      {
         vn=(*vl).v[i]+j;

         if ((*vn).type!=EXTERNAL)
         {
            for (k=0;k<(int)(LTYPES);k++)
            {
               ln=rvt[(ltype_t)(k)];

               if (((*vn).nl[(int)(ln)]>0)&&((*v).nl[k]>(k==(int)(l))))
                  nup+=1;
            }
         }
      }
   }

   if (nup>0)
   {
      up=malloc(nup*sizeof(*up));
      error(up==NULL,1,"add_vertices [branch.c]",
            "Unable to add further vertices");
      (*bv).nup=nup;
      (*bv).up=up;

      for (i=0;i<=imx;i++)
      {
         for (j=0;j<(*vl).nv[i];j++)
         {
            vn=(*vl).v[i]+j;

            if ((*vn).type!=EXTERNAL)
            {
               for (k=0;k<(int)(LTYPES);k++)
               {
                  ln=rvt[(ltype_t)(k)];

                  if (((*vn).nl[(int)(ln)]>0)&&((*v).nl[k]>(k==(int)(l))))
                  {
                     (*up).order=(*bv).order+i;
                     (*up).ibr=nbr;
                     (*up).v[0]=i;
                     (*up).v[1]=j;
                     (*up).l=(ltype_t)(k);

                     (*up).down=bv;
                     (*up).nup=0;
                     (*up).up=NULL;

                     up+=1;
                     nbr+=1;
                  }
               }
            }
         }
      }

      for (i=0;i<(*bv).nup;i++)
         nbr=add_vertices(order,nbr,vl,(*bv).up+i);
   }

   return nbr;
}


static void bv2btag(bvertex_t *bv,bvertex_t **bt)
{
   int ibr,nup,i;
   bvertex_t *up;

   ibr=(*bv).ibr;
   bt[ibr]=bv;

   nup=(*bv).nup;
   up=(*bv).up;

   for (i=0;i<nup;i++)
      bv2btag(up+i,bt);
}


static void set_btag(theory_t th)
{
   int nbr,nup,i;
   bvertex_t **bt,*up;

   nbr=bls[th].nbr;
   nup=bls[th].nup;
   up=bls[th].up;

   bt=malloc(nbr*sizeof(*bt));
   error(bt==NULL,1,"set_btag [branch.c]",
         "Unable to allocate branch tag array");

   for (i=0;i<nup;i++)
      bv2btag(up+i,bt);

   btag[th]=bt;
}


static void set_binv(theory_t th)
{
   int nbr,nvm,ibr;
   int nv,i,j,*bi;
   branch_t *br1,*br2;

   nbr=bls[th].nbr;
   nvm=bls[th].order+2;

   bi=malloc(nbr*sizeof(*bi));
   error(bi==NULL,1,"set_binv [branch.c]",
         "Unable to allocate index array");

   br1=alloc_branch(nvm);
   br2=alloc_branch(nvm);

   for (ibr=0;ibr<nbr;ibr++)
   {
      (*br1).nv=nvm;
      get_branch(th,ibr,br1);
      nv=(*br1).nv;
      (*br2).nv=nv;

      for (i=0;i<nv;i++)
      {
         j=nv-i-1;
         (*br2).v[i][0]=(*br1).v[j][0];
         (*br2).v[i][1]=(*br1).v[j][1];

         if (i<(nv-1))
            (*br2).l[i]=rvt[(int)((*br1).l[j-1])];
      }

      bi[ibr]=find_branch(th,br2);
   }

   free_branch(br1);
   free_branch(br2);

   binv[(int)(th)]=bi;
}


int branches(theory_t th,int order)
{
   int nbr,nup,imx,i,j;
   vertex_t *vn;
   bvertex_t *up;
   blist_t *bl;
   vlist_t vl;

   error((th==THEORIES)||(order<0),1,"branches [branch.c]",
         "Arguments are out of range");

   bl=bls+(int)(th);
   nbr=(*bl).nbr;

   if (nbr==0)
   {
      vl=vertices(th,order);
      imx=order;
      if (imx>vl.order)
         imx=vl.order;
      nup=0;

      for (i=0;i<=imx;i++)
      {
         for (j=0;j<vl.nv[i];j++)
         {
            vn=vl.v[i]+j;

            if ((*vn).type!=EXTERNAL)
               nup+=1;
         }
      }

      up=malloc(nup*sizeof(*up));
      error(up==NULL,1,"branches [branch.c]",
            "Unable to allocate initial branch array");

      (*bl).order=order;
      (*bl).nup=nup;
      (*bl).up=up;

      for (i=0;i<=imx;i++)
      {
         for (j=0;j<vl.nv[i];j++)
         {
            vn=vl.v[i]+j;

            if ((*vn).type!=EXTERNAL)
            {
               (*up).order=i;
               (*up).ibr=nbr;
               (*up).v[0]=i;
               (*up).v[1]=j;
               (*up).l=LTYPES;

               (*up).down=NULL;
               (*up).nup=0;
               (*up).up=NULL;

               nbr+=1;
               up+=1;
            }
         }
      }

      for (i=0;i<(*bl).nup;i++)
         nbr=add_vertices(order,nbr,&vl,(*bl).up+i);

      (*bl).nbr=nbr;
      set_btag(th);
      set_binv(th);
   }
   else
      error(order!=(*bl).order,1,"branches [branch.c]",
            "Attempt to modify the branch list");

   return nbr;
}


int get_nv(theory_t th,int ibr)
{
   int nv;
   bvertex_t *bv;

   error((th==THEORIES)||(ibr<0)||(ibr>=bls[th].nbr),1,"get_nv [branch.c]",
         "Arguments are out of range");

   bv=btag[th][ibr];
   nv=0;

   while (bv!=NULL)
   {
      nv+=1;
      bv=(*bv).down;
   }

   return nv;
}


int get_order(theory_t th,int ibr)
{
   bvertex_t *bv;

   error((th==THEORIES)||(ibr<0)||(ibr>=bls[th].nbr),1,"get_order [branch.c]",
         "Arguments are out of range");

   bv=btag[th][ibr];

   return (*bv).order;
}


void get_branch(theory_t th,int ibr,branch_t *br)
{
   int nv,i;
   int (*v)[2];
   ltype_t *l;
   bvertex_t *bv;

   nv=get_nv(th,ibr);

   error((*br).nv<nv,1,"get_branch [branch.c]",
         "Arrays in branch structure are too short");

   v=(*br).v;
   l=(*br).l;
   bv=btag[th][ibr];

   for (i=(nv-1);i>=0;i--)
   {
      v[i][0]=(*bv).v[0];
      v[i][1]=(*bv).v[1];
      if (i>0)
         l[i-1]=(*bv).l;
      bv=(*bv).down;
   }

   (*br).nv=nv;
}


int find_branch(theory_t th,branch_t *br)
{
   int nv,nup,i,j;
   int (*v)[2];
   ltype_t *l;
   bvertex_t *bv,*up;

   error((th==THEORIES)||(bls[th].nbr==0),1,
         "find_branch [branch.c]","Unknown theory or missing branch list");

   nv=(*br).nv;
   v=(*br).v;
   l=(*br).l;
   bv=NULL;

   for (i=0;i<nv;i++)
   {
      if (i==0)
         nup=bls[th].nup;
      else
         nup=(*bv).nup;

      for (j=0;j<nup;j++)
      {
         if (i==0)
            up=bls[th].up+j;
         else
            up=(*bv).up+j;

         if (((*up).v[0]==v[i][0])&&((*up).v[1]==v[i][1]))
         {
            if ((i==0)||((*up).l==l[i-1]))
            {
               if (i==(nv-1))
                  return (*up).ibr;
               else
               {
                  bv=up;
                  break;
               }
            }
         }
      }

      if (j==nup)
         error(1,1,"find_branch [branch.c]","Branch not found");
   }

   return 0;
}


int inverse_branch(theory_t th,int ibr)
{
   error((th==THEORIES)||(ibr<0)||(ibr>=bls[th].nbr),1,
         "inverse_branch [branch.c]","Arguments are out of range");

   return binv[(int)(th)][ibr];
}


void branch2dgm(branch_t *br,diagram_t *dm)
{
   int nv,nl,i;
   int (*v1)[2],(*v2)[2];
   ltype_t *l1;
   line_t *l2;

   nv=(*br).nv;
   nl=nv-1;

   error((nv>(*dm).nv)||(nl>(*dm).nl),1,"branch2dgm [branch.c]",
         "Arrays in diagram structure are too short");

   (*dm).nv=nv;
   (*dm).nl=nl;

   v1=(*br).v;
   v2=(*dm).v;

   for (i=0;i<nv;i++)
   {
      (*v2)[0]=(*v1)[0];
      (*v2)[1]=(*v1)[1];

      v1+=1;
      v2+=1;
   }

   l1=(*br).l;
   l2=(*dm).l;

   for (i=0;i<nl;i++)
   {
      (*l2).type=(*l1);
      (*l2).iv[0]=i;
      (*l2).iv[1]=i+1;

      l1+=1;
      l2+=1;
   }
}
